// "C64 Emulator V1.0". 
// Programmed for TI-92+ by Sebastian Macke, 
// You need TIGCC crosscompiler to compile it.

#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization

#define SAVE_SCREEN           // Save/Restore LCD Contents

#include <tigcclib.h>         // Include All Header Files
#include "cpuemu.h"

short _ti92plus;              // Produce .9XZ File

unsigned char *c64mem, *c64chars,*LCD;

void draw_string(int y,unsigned char *s)
{
	int i;
	int pos;
	int offset;
	offset = y*30;
	pos = 0;

  for(i=0;i<=9;i++)
  {
		LCD[offset]     = c64chars[s[pos]*5];
		LCD[offset+30]  = c64chars[s[pos]*5+1];
		LCD[offset+60]  = c64chars[s[pos]*5+2];
		LCD[offset+90]  = c64chars[s[pos]*5+3];
		LCD[offset+120] = c64chars[s[pos]*5+4];
		pos++;
	 
		LCD[offset]     += ((c64chars[s[pos]*5]   >> 6) & 3);
		LCD[offset+30]  += ((c64chars[s[pos]*5+1] >> 6) & 3);
		LCD[offset+60]  += ((c64chars[s[pos]*5+2] >> 6) & 3);
		LCD[offset+90]  += ((c64chars[s[pos]*5+3] >> 6) & 3);
		LCD[offset+120] += ((c64chars[s[pos]*5+4] >> 6) & 3);
		offset++;
	
		LCD[offset]     = ((c64chars[s[pos]*5] << 2) );
		LCD[offset+30]  = ((c64chars[s[pos]*5+1] << 2) );
		LCD[offset+60]  = ((c64chars[s[pos]*5+2] << 2) );
		LCD[offset+90]  = ((c64chars[s[pos]*5+3] << 2) );
		LCD[offset+120] = ((c64chars[s[pos]*5+4] << 2) );

		pos++;

		LCD[offset]     += ((c64chars[s[pos]*5]   >> 4) & 15);
		LCD[offset+30]  += ((c64chars[s[pos]*5+1] >> 4) & 15);
		LCD[offset+60]  += ((c64chars[s[pos]*5+2] >> 4) & 15);
		LCD[offset+90]  += ((c64chars[s[pos]*5+3] >> 4) & 15);
		LCD[offset+120] += ((c64chars[s[pos]*5+4] >> 4) & 15);

		offset++;	
		LCD[offset]     = ((c64chars[s[pos]*5] << 4));
		LCD[offset+30]  = ((c64chars[s[pos]*5+1] << 4));
		LCD[offset+60]  = ((c64chars[s[pos]*5+2] << 4));
		LCD[offset+90]  = ((c64chars[s[pos]*5+3] << 4));
		LCD[offset+120] = ((c64chars[s[pos]*5+4] << 4));

  	pos++;
		LCD[offset]     += ((c64chars[s[pos]*5] >> 2)   & 63);
		LCD[offset+30]  += ((c64chars[s[pos]*5+1] >> 2) & 63);
		LCD[offset+60]  += ((c64chars[s[pos]*5+2] >> 2) & 63);
		LCD[offset+90]  += ((c64chars[s[pos]*5+3] >> 2) & 63);
		LCD[offset+120] += ((c64chars[s[pos]*5+4] >> 2) & 63);

		pos++;
		offset++;
  }

}

void draw_screen()
{
	int i;
	for(i=0;i<=24;i++) 
		draw_string(i*5,&c64mem[1024+i*40]);  
	for(i=3750;i<=3840;i++)
		LCD[i]=0; 
}

unsigned char *membuffer;

// this is a very bad solution, because it writes behind the known ram 
// in the hope, that there is no important data.
// if somebody knows a better solution eMail me: sebastian@macke.de
void mem_on(BOOL x)
{
	static unsigned char buffer[20];
	unsigned int i;
	if (x == TRUE)
	{
		for(i=0;i<=19;i++) buffer[i] = c64mem[0xFFEC+i];
		for(i=0;i<=19;i++) c64mem[0xFFEC+i] = membuffer[i];
	}
	else
	{
		for(i=0;i<=19;i++) membuffer[i] = c64mem[0xFFEC+i];
		for(i=0;i<=19;i++) c64mem[0xFFEC+i] = buffer[i];
	}
}


void  rom2ram(char *rom)
{
	unsigned int i;

	for(i=0;i<=8191;i++) 
	{
		c64mem[0xe000+i] = rom[i];
		c64mem[0xa000+i] = rom[8192+i];
	}
}


void calc_steps(unsigned int NUM)
{
	unsigned int i;

	for(i=0;i<=NUM-1;i++) 
	{
		calc_step();
	}

}

// Main Function
void _main(void)
{ 
  unsigned int i;
  short key;
  SYM_ENTRY *SYM;
	
	
	key=0;
	clrscr ();
  LCD = LCD_MEM;

	SYM = SymFindPtr ($(c64chars), 0);
  c64chars = HeapDeref (SYM->handle); 
  c64chars+=2;

	SYM = SymFindPtr ($(c64mem), 0);
	c64mem = HeapDeref (SYM->handle); 
  c64mem+=2;

	SYM = SymFindPtr ($(c64rom), 0);
	membuffer = HeapDeref (SYM->handle); 
  membuffer+=102;
  membuffer+=(8192+8192);

	init(c64mem,(struct tcpuregisters*) (membuffer-8192-8192-50));
	draw_screen();
	DlgMessage ("programmed by Sebastian Macke", "C64 Emulator for TI92P", BT_NONE, BT_NONE); 
	if (kbhit ()) key = ngetchx ();
			
	mem_on(TRUE);
	do
  {
		calc_steps(2000);
		draw_screen();
		interrupt();

		if (kbhit ()) key = ngetchx ();
	
		if (key == KEY_MODE) 
		{	
  		resetcpu();
  		rom2ram(membuffer-8192-8192);
  		set_PC(0xe394);
	
			for(i=0;i<=40*25-1;i++) 
			c64mem[1024+i] = 32;
			key=0;	
		}

		
  } while (key != KEY_APPS);
	mem_on(FALSE);

}
