// "C64 Emulator V1.0". 
// Programmed for TI-92+ by Sebastian Macke, 
// You need TIGCC crosscompiler to compile it.

#include <string.h>
#include <kbd.h>
#include <system.h>
#include "cpuemu.h"

#define OP_NONE		0
#define OP_ADC		1
#define OP_AND		2
#define OP_ASL		3
#define OP_BCC		4
#define OP_BCS		5
#define OP_BEQ		6
#define OP_BIT		7
#define OP_BMI		8
#define OP_BNE		9
#define OP_BPL		10
#define OP_BRK		11
#define OP_BVC		12
#define OP_BVS		13
#define OP_CLC		14
#define OP_CLD		15	
#define OP_CLI		16
#define OP_CLV		17
#define OP_CMP		18
#define OP_CPX		19
#define OP_CPY		20
#define OP_DEC		21
#define OP_DEX		22
#define OP_DEY		23
#define OP_EOR		24
#define OP_INC		25
#define OP_INX		26
#define OP_INY		27
#define OP_JMP		28
#define OP_JSR		29
#define OP_LDA		30
#define OP_LDX		31
#define OP_LDY		32
#define OP_LSR		33
#define OP_NOP		34
#define OP_ORA		35
#define OP_PHA		36
#define OP_PLA		37
#define OP_PHP		38
#define OP_PLP		39
#define OP_ROL		40
#define OP_ROR		41
#define OP_RTI		42
#define OP_RTS		43
#define OP_SBC		44
#define OP_SEC		45
#define OP_SED		46
#define OP_SEI		47
#define OP_STA		48
#define OP_STX		49
#define OP_STY		50
#define OP_TAX		51
#define OP_TAY		52
#define OP_TSX		53
#define OP_TXA		54
#define OP_TXS		55
#define OP_TYA		56

#define AMODE_IMP  0
#define AMODE_ACC  1
#define AMODE_ABS  2
#define AMODE_ZP   3
#define AMODE_IMM  4
#define AMODE_ABSX 5
#define AMODE_ABSY 6
#define AMODE_INDX 7
#define AMODE_INDY 8
#define AMODE_ZPX  9
#define AMODE_ZPY  10
#define AMODE_REL  11
#define AMODE_IND  12

struct tkeys
{
  unsigned char tirow,tibit,c64row,c64bit,shift;
};

struct tcode2asm
{
  unsigned char ass[4],amode, op;
};

struct tcode2asm opcode[]={
"BRK",AMODE_IMP,OP_BRK,
"ORA",AMODE_INDX,OP_ORA,
"non",0,0,
"non",0,0,
"non",0,0,
"ORA",AMODE_ZP,OP_ORA,
"ASL",AMODE_ZP,OP_ASL,
"non",0,0,
"PHP",AMODE_IMP,OP_PHP,
"ORA",AMODE_IMM,OP_ORA,
"ASL",AMODE_ACC,OP_ASL,
"non",0,0,
"non",0,0,
"ORA",AMODE_ABS,OP_ORA,
"ASL",AMODE_ABS,OP_ASL,
"non",0,0,
"BPL",AMODE_REL,OP_BPL,
"ORA",AMODE_INDY,OP_ORA,
"non",0,0,
"non",0,0,
"non",0,0,
"ORA",AMODE_ZPX,OP_ORA,
"ASL",AMODE_ZPX,OP_ASL,
"non",0,0,
"CLC",AMODE_IMP,OP_CLC,
"ORA",AMODE_ABSY,OP_ORA,
"non",0,0,
"non",0,0,
"non",0,0,
"ORA",AMODE_ABSX,OP_ORA,
"ASL",AMODE_ABSX,OP_ASL,
"non",0,0,
"JSR",AMODE_ABS,OP_JSR,
"AND",AMODE_INDX,OP_AND,
"non",0,0,
"non",0,0,
"BIT",AMODE_ZP,OP_BIT,
"AND",AMODE_ZP,OP_AND,
"ROL",AMODE_ZP,OP_ROL,
"non",0,0,
"PLP",AMODE_IMP,OP_PLP,
"AND",AMODE_IMM,OP_AND,
"ROL",AMODE_ACC,OP_ROL,
"non",0,0,
"BIT",AMODE_ABS,OP_BIT,
"AND",AMODE_ABS,OP_AND,
"ROL",AMODE_ABS,OP_ROL,
"non",0,0,
"BMI",AMODE_REL,OP_BMI,
"AND",AMODE_INDY,OP_AND,
"non",0,0,
"non",0,0,
"non",0,0,
"AND",AMODE_ZPX,OP_AND,
"ROL",AMODE_ZPX,OP_ROL,
"non",0,0,
"SEC",AMODE_IMP,OP_SEC,
"AND",AMODE_ABSY,OP_AND,
"non",0,0,
"non",0,0,
"non",0,0,
"AND",AMODE_ABSX,OP_AND,
"ROL",AMODE_ABSX,OP_ROL,
"non",0,0,
"RTI",AMODE_IMP,OP_RTI,
"EOR",AMODE_INDX,OP_EOR,
"non",0,0,
"non",0,0,
"non",0,0,
"EOR",AMODE_ZP,OP_EOR,
"LSR",AMODE_ZP,OP_LSR,
"non",0,0,
"PHA",AMODE_IMP,OP_PHA,
"EOR",AMODE_IMM,OP_EOR,
"LSR",AMODE_ACC,OP_LSR,
"non",0,0,
"JMP",AMODE_ABS,OP_JMP,
"EOR",AMODE_ABS,OP_EOR,
"LSR",AMODE_ABS,OP_LSR,
"non",0,0,
"BVC",AMODE_REL,OP_BVC,
"EOR",AMODE_INDY,OP_EOR,
"non",0,0,
"non",0,0,
"non",0,0,
"EOR",AMODE_ZPX,OP_EOR,
"LSR",AMODE_ZPX,OP_LSR,
"non",0,0,
"CLI",AMODE_IMP,OP_CLI,
"EOR",AMODE_ABSY,OP_EOR,
"non",0,0,
"non",0,0,
"non",0,0,
"EOR",AMODE_ABSX,OP_EOR,
"LSR",AMODE_ABSX,OP_LSR,
"non",0,0,
"RTS",AMODE_IMP,OP_RTS,
"ADC",AMODE_INDX,OP_ADC,
"non",0,0,
"non",0,0,
"non",0,0,
"ADC",AMODE_ZP,OP_ADC,
"ROR",AMODE_ZP,OP_ROR,
"non",0,0,
"PLA",AMODE_IMP,OP_PLA,
"ADC",AMODE_IMM,OP_ADC,
"ROR",AMODE_ACC,OP_ROR,
"non",0,0,
"JMP",AMODE_IND,OP_JMP,
"ADC",AMODE_ABS,OP_ADC,
"ROR",AMODE_ABS,OP_ROR,
"non",0,0,
"BVS",AMODE_REL,OP_BVS,
"ADC",AMODE_INDY,OP_ADC,
"non",0,0,
"non",0,0,
"non",0,0,
"ADC",AMODE_ZPX,OP_ADC,
"ROR",AMODE_ZPX,OP_ROR,
"non",0,0,
"SEI",AMODE_IMP,OP_SEI,
"ADC",AMODE_ABSY,OP_ADC,
"non",0,0,
"non",0,0,
"non",0,0,
"ADC",AMODE_ABSX,OP_ADC,
"ROR",AMODE_ABSX,OP_ROR,
"non",0,0,
"NOP",AMODE_IMP,OP_NOP,
"STA",AMODE_INDX,OP_STA,
"NOP",AMODE_IMP,OP_NOP,
"non",0,0,
"STY",AMODE_ZP,OP_STY,
"STA",AMODE_ZP,OP_STA,
"STX",AMODE_ZP,OP_STX,
"non",0,0,
"DEY",AMODE_IMP,OP_DEY,
"NOP",AMODE_IMP,OP_NOP,
"TXA",AMODE_IMP,OP_TXA,
"non",0,0,
"STY",AMODE_ABS,OP_STY,
"STA",AMODE_ABS,OP_STA,
"STX",AMODE_ABS,OP_STX,
"non",0,0,
"BCC",AMODE_REL,OP_BCC,
"STA",AMODE_INDY,OP_STA,
"non",0,0,
"non",0,0,
"STY",AMODE_ZPX,OP_STY,
"STA",AMODE_ZPX,OP_STA,
"STX",AMODE_ZPY,OP_STX,
"non",0,0,
"TYA",AMODE_IMP,OP_TYA,
"STA",AMODE_ABSY,OP_STA,
"TXS",AMODE_IMP,OP_TXS,
"non",0,0,
"non",0,0,
"STA",AMODE_ABSX,OP_STA,
"non",0,0,
"non",0,0,
"LDY",AMODE_IMM,OP_LDY,
"LDA",AMODE_INDX,OP_LDA,
"LDX",AMODE_IMM,OP_LDX,
"non",0,0,
"LDY",AMODE_ZP,OP_LDY,
"LDA",AMODE_ZP,OP_LDA,
"LDX",AMODE_ZP,OP_LDX,
"non",0,0,
"TAY",AMODE_IMP,OP_TAY,
"LDA",AMODE_IMM,OP_LDA,
"TAX",AMODE_IMP,OP_TAX,
"non",0,0,
"LDY",AMODE_ABS,OP_LDY,
"LDA",AMODE_ABS,OP_LDA,
"LDX",AMODE_ABS,OP_LDX,
"non",0,0,
"BCS",AMODE_REL,OP_BCS,
"LDA",AMODE_INDY,OP_LDA,
"non",0,0,
"non",0,0,
"LDY",AMODE_ZPX,OP_LDY,
"LDA",AMODE_ZPX,OP_LDA,
"LDX",AMODE_ZPY,OP_LDX,
"non",0,0,
"CLD",AMODE_IMP,OP_CLD,
"LDA",AMODE_ABSY,OP_LDA,
"TSX",AMODE_IMP,OP_TSX,
"non",0,0,
"LDY",AMODE_ABSX,OP_LDY,
"LDA",AMODE_ABSX,OP_LDA,
"LDX",AMODE_ABSY,OP_LDX,
"non",0,0,
"CPY",AMODE_IMM,OP_CPY,
"CMP",AMODE_INDX,OP_CMP,
"NOP",AMODE_IMP,OP_NOP,
"non",0,0,
"CPY",AMODE_ZP,OP_CPY,
"CMP",AMODE_ZP,OP_CMP,
"DEC",AMODE_ZP,OP_DEC,
"non",0,0,
"INY",AMODE_IMP,OP_INY,
"CMP",AMODE_IMM,OP_CMP,
"DEX",AMODE_IMP,OP_DEX,
"non",0,0,
"CPY",AMODE_ABS,OP_CPY,
"CMP",AMODE_ABS,OP_CMP,
"DEC",AMODE_ABS,OP_DEC,
"non",0,0,
"BNE",AMODE_REL,OP_BNE,
"CMP",AMODE_INDY,OP_CMP,
"non",0,0,
"non",0,0,
"non",0,0,
"CMP",AMODE_ZPX,OP_CMP,
"DEC",AMODE_ZPX,OP_DEC,
"non",0,0,
"CLV",AMODE_IMP,OP_CLV,
"CMP",AMODE_ABSY,OP_CMP,
"non",0,0,
"non",0,0,
"non",0,0,
"CMP",AMODE_ABSX,OP_CMP,
"DEC",AMODE_ABSX,OP_DEC,
"non",0,0,
"CPX",AMODE_IMM,OP_CPX,
"SBC",AMODE_INDX,OP_SBC,
"NOP",AMODE_IMP,OP_NOP,
"non",0,0,
"CPX",AMODE_ZP,OP_CPX,
"SBC",AMODE_ZP,OP_SBC,
"INC",AMODE_ZP,OP_INC,
"non",0,0,
"INX",AMODE_IMP,OP_INX,
"SBC",AMODE_IMM,OP_SBC,
"NOP",AMODE_IMP,OP_NOP,
"non",0,0,
"CPX",AMODE_ABS,OP_CPX,
"SBC",AMODE_ABS,OP_SBC,
"INC",AMODE_ABS,OP_INC,
"non",0,0,
"BEQ",AMODE_REL,OP_BEQ,
"SBC",AMODE_INDY,OP_SBC,
"non",0,0,
"non",0,0,
"non",0,0,
"SBC",AMODE_ZPX,OP_SBC,
"INC",AMODE_ZPX,OP_INC,
"non",0,0,
"SED",AMODE_IMP,OP_SED,
"SBC",AMODE_ABSY,OP_SBC,
"non",0,0,
"non",0,0,
"non",0,0,
"SBC",AMODE_ABSX,OP_SBC,
"INC",AMODE_ABSX,OP_INC,
"non",0,0
};

struct tkeys keys[]={
	0,7,0,7,FALSE,			//down
	0,6,0,2,FALSE,			//right
	0,5,0,7,TRUE,				//up
	0,4,0,2,TRUE,				//left

	6,4,0,4,FALSE,			//F1
	2,4,0,5,FALSE,			//F3 
	7,4,0,6,FALSE,			//F5
	3,4,0,3,FALSE,			//F7
	4,4,0,4,TRUE,  			//F2
	9,4,0,5,TRUE,	  		//F4
	5,4,0,6,TRUE,		  	//F6
	1,4,0,3,TRUE,			  //F8

	5,7,6,3,FALSE,			//HOME
	6,0,6,0,FALSE,			//LIRA
	

	0,1,7,5,FALSE,			//C=
	8,6,7,7,FALSE,			//STOP
	8,4,5,0,FALSE,			//+
	9,0,5,3,FALSE,			//-
	7,7,6,1,FALSE,			//*
	5,0,6,7,FALSE,			// /
	7,0,6,5,FALSE,			//=
	5,5,5,5,FALSE,			//:
	5,6,6,2,FALSE,			//;
	4,7,5,7,FALSE,			//,
	9,6,5,4,FALSE,			//.

	4,5,3,3,TRUE,	  		//(
	4,6,4,0,TRUE,		  	//)

	9,1,0,1,FALSE,			//Enter
	6,6,0,1,FALSE,			//Enter

	0,2,6,4,FALSE,			//Shift
	4,0,7,4,FALSE,			//Space
	8,0,0,0,FALSE,			//Back
	6,5,7,1,FALSE,			//<-

	9,5,4,3,FALSE,			//0
	1,5,7,0,FALSE,			//1
	1,6,7,3,FALSE,			//2
	1,7,1,0,FALSE,			//3
	2,5,1,3,FALSE,			//4
	2,6,2,0,FALSE,			//5
	2,7,2,3,FALSE,			//6
	3,5,3,0,FALSE,			//7
	3,6,3,3,FALSE,			//8
	3,7,4,0,FALSE,			//9

	9,2,1,2,FALSE,			//a
	5,1,3,4,FALSE,			//b
	3,1,2,4,FALSE,			//c
	2,2,2,2,FALSE,			//d
	2,3,1,6,FALSE,			//e
	3,2,2,5,FALSE,			//f
	4,2,3,2,FALSE,			//g
	5,2,3,5,FALSE,			//h
	7,3,4,1,FALSE,			//i
	6,2,4,2,FALSE,			//j
	7,2,4,5,FALSE,			//k
	8,2,5,2,FALSE,			//l
	7,1,4,4,FALSE,			//m
	6,1,4,7,FALSE,			//n
	8,3,4,6,FALSE,			//o
	6,7,5,1,FALSE,			//p
	9,3,7,6,FALSE,			//q
	3,3,2,1,FALSE,			//r
	1,2,1,5,FALSE,			//s
	4,3,2,6,FALSE,			//t
	6,3,3,6,FALSE,			//u
	4,1,3,7,FALSE,			//v
	1,3,1,1,FALSE,			//w
	2,1,2,7,FALSE,			//x
	5,3,3,1,FALSE,			//y
	1,1,1,4,FALSE				//z
};


struct tcpuregisters* regs={0};
unsigned char* mem={0};

void resetcpu()
{
	regs->AC = 0;
	regs->XR = 0;
	regs->YR = 0;
	regs->PC = 0;
	regs->SP = 0xFF;
	regs->Carry_Flag = 0;
	regs->Zero_Flag = 0;
	regs->IRQ_Disable_Flag = 0;
	regs->Decimal_Flag = 0;
	regs->Brk_Flag = 0;
	regs->Unused_Flag = 1;
	regs->Overflow_Flag = 0;
	regs->Negative_Flag = 0;
}

void init(unsigned char *pmem,struct tcpuregisters *pregs)
{
	mem = pmem;
	regs = pregs;
}


void set_PC(unsigned int pos)
{
	regs->PC = pos;
}

unsigned char getstatus()
{
	unsigned char x;
	x = 0;
	if (regs->Carry_Flag) x = x | 1;
	if (regs->Zero_Flag)  x = x | 2;
	if (regs->IRQ_Disable_Flag)  x = x | 4;
	if (regs->Decimal_Flag) x = x | 8;
	if (regs->Brk_Flag) x = x | 16;
	x = x | 32;
	if (regs->Overflow_Flag) x = x | 64;
	if (regs->Negative_Flag) x = x | 128;
	return x;
}

void putstatus(unsigned char x)
{
	regs->Carry_Flag = FALSE;
	regs->Zero_Flag = FALSE;
	regs->IRQ_Disable_Flag = FALSE;
	regs->Decimal_Flag = FALSE;
	regs->Brk_Flag = FALSE;
	regs->Unused_Flag = TRUE;
	regs->Overflow_Flag = FALSE;
	regs->Negative_Flag = FALSE;
	if (x & 1) regs->Carry_Flag = TRUE;
	if (x & 2) regs->Zero_Flag = TRUE;
	if (x & 4) regs->IRQ_Disable_Flag = TRUE;
	if (x & 8) regs->Decimal_Flag = TRUE;
	if (x & 16) regs->Brk_Flag = TRUE;
	if (x & 64) regs->Overflow_Flag = TRUE;
	if (x & 128) regs->Negative_Flag = TRUE;
}

void print_processor_status()
{
	unsigned char address_mode, x;
	unsigned int address;
	unsigned char s[20];


	printf("AC: %u\n",regs->AC);

	printf("XR: %u YR: %u\n",regs->XR, regs->YR);
	printf("PC: %u SP: %u\n",regs->PC, regs->SP);

	printf("Carry Flag: %u\n",regs->Carry_Flag);
	printf("Zero Flag: %u\n",regs->Zero_Flag);
	printf("Overflow Flag: %u\n",regs->Overflow_Flag);
	printf("Negative Flag: %u\n",regs->Negative_Flag);

	address_mode = opcode[mem[regs->PC]].amode;

	switch (address_mode)
	{
		case AMODE_IMP:
      sprintf(s,"%s\n",opcode[mem[regs->PC]].ass);
      break;

		case AMODE_ACC:
      sprintf(s,"%s A\n",opcode[mem[regs->PC]].ass);
      break;

		case AMODE_ABS:
      address = ((unsigned int)mem[regs->PC+2] << 8) | mem[regs->PC+1];
      sprintf(s,"%s %u\n",opcode[mem[regs->PC]].ass,address);
      break;

		case AMODE_IMM:
      x = mem[regs->PC+1];
      sprintf(s,"%s #%u\n",opcode[mem[regs->PC]].ass,x);
      break;

		case AMODE_ZP:
      address = mem[regs->PC+1];
      sprintf(s,"%s %u\n",opcode[mem[regs->PC]].ass,address);
      break;

		case AMODE_ABSX:
      address = ((unsigned int)mem[regs->PC+2] << 8) + mem[regs->PC+1];
      sprintf(s,"%s %u,x\n",opcode[mem[regs->PC]].ass,address);
      break;

		case AMODE_ABSY:
      address = ((unsigned int)mem[regs->PC+2] << 8) + mem[regs->PC+1];
      sprintf(s,"%s %u,y\n",opcode[mem[regs->PC]].ass,address);
      break;

		case AMODE_ZPX:
      address = (mem[regs->PC+1]+regs->XR) & 0xFF;
      sprintf(s,"%s %u,x\n",opcode[mem[regs->PC]].ass,address);
      break;

		case AMODE_ZPY:
      address = (mem[regs->PC+1]+regs->YR) & 0xFF;
      sprintf(s,"%s %u,y\n",opcode[mem[regs->PC]].ass,address);
      break;

		case AMODE_REL:
      x = mem[regs->PC+1];
      sprintf(s,"%s %d\n",opcode[mem[regs->PC]].ass,(signed char)x);
      break;

		case AMODE_IND:
      address = ((unsigned int)mem[regs->PC+2] << 8) | mem[regs->PC+1];
      sprintf(s,"%s (%u)\n",opcode[mem[regs->PC]].ass,address);
      break;

		case AMODE_INDX:
      address = (mem[regs->PC+1] + regs->XR) & 0xFF;
      sprintf(s,"%s (%u,x)\n",opcode[mem[regs->PC]].ass,address);
      break;

		case AMODE_INDY:
      address = mem[regs->PC+1];
      sprintf(s,"%s (%u),y\n",opcode[mem[regs->PC]].ass,address);
      break;
  
	}

	printf("next command: %s\n",s);

}

void	interrupt()
{
	mem[256+regs->SP] = ((regs->PC) >> 8) & 0xFF;
	regs->SP--;
	mem[256+regs->SP] = (regs->PC) & 0xFF;
	regs->SP--;
	regs->Brk_Flag = 0;
	mem[256+regs->SP] = getstatus();
	regs->SP--;
	regs->IRQ_Disable_Flag = 1;
  regs->PC = (unsigned int)mem[0xFFFE] | ((unsigned int)mem[0xFFFF] << 8);
}

char getkey(unsigned int x)
{
	unsigned char tikeys[10];
	unsigned int shift;
	unsigned int i;
	unsigned char ret; 

	ret = 0;
	shift=0;
	OSSetSR (0x0700);
	tikeys[0] = _rowread(~1);
	tikeys[1] = _rowread(~2);
	tikeys[2] = _rowread(~4);
	tikeys[3] = _rowread(~8);
	tikeys[4] = _rowread(~16);
	tikeys[5] = _rowread(~32);
	tikeys[6] = _rowread(~64);
	tikeys[7] = _rowread(~128);
	tikeys[8] = _rowread(~256);
	tikeys[9] = _rowread(~512);
	OSSetSR (0x0000);

	for (i=0;i<=68;i++)
	{
		if (tikeys[keys[i].tirow] & (1<<keys[i].tibit))
		{
			if (keys[i].shift) shift=1;
			if ((x & (1<<keys[i].c64row))==0) ret = ret | (1<<keys[i].c64bit);
		}

		if((shift) && ((x & 2) == 0)) ret = ret | (1<<7);
	}
	return (~ret);
}


void calc_step()
{
	unsigned char command, address_mode;
	unsigned char x;
	unsigned int address;
	unsigned char code;
	unsigned char stack; //only used by PLP and  RTI
	unsigned int help, help2; //only in some operations used like adc and cmp

	address = 0;
	x=0;

	command = mem[regs->PC];
	address_mode = opcode[command].amode;
	code = opcode[command].op;

	switch (address_mode)
	{
		case AMODE_IMP:
     	regs->PC += 1;
      break;
		
		case AMODE_ACC:
      x = regs->AC;
     	regs->PC += 1;
      break;

		case AMODE_ABS:
      address = ((unsigned int)mem[regs->PC+2] << 8) | mem[regs->PC+1];
      x = mem[address];
     	regs->PC += 3;
      break;

		case AMODE_IMM:
      x = mem[regs->PC+1];
     	regs->PC += 2;
      break;

		case AMODE_ZP:
      address = mem[regs->PC+1];
			x = mem[address];      
     	regs->PC += 2;
      break;

		case AMODE_ABSX:
      address = (((unsigned int)mem[regs->PC+2] << 8) | mem[regs->PC+1])+regs->XR;
      x = mem[address];
     	regs->PC += 3;
      break;

		case AMODE_ABSY:
      address = (((unsigned int)mem[regs->PC+2] << 8) | mem[regs->PC+1])+regs->YR;
      x = mem[address];
     	regs->PC += 3;
      break;

		case AMODE_ZPX:
      address = (mem[regs->PC+1]+regs->XR) & 0xFF;
      x = mem[address];
     	regs->PC += 2;
      break;

		case AMODE_ZPY:
      address = (mem[regs->PC+1]+regs->YR) & 0xFF;
      x = mem[address];
     	regs->PC += 2;
      break;

		case AMODE_REL:
      x = mem[regs->PC+1];
     	regs->PC += 2;
      break;

		case AMODE_IND:
      address = ((unsigned int)mem[regs->PC+2] << 8) | mem[regs->PC+1];
      address = ((unsigned int)mem[address+1] << 8) | mem[address];
     	regs->PC += 3;
      break;

		case AMODE_INDX:
      address = (mem[regs->PC+1] + regs->XR) & 0xFF;
      address = ((unsigned int)mem[address+1] << 8) | mem[address];
      x = mem[address];
     	regs->PC += 2;
      break;

		case AMODE_INDY:
      address = mem[regs->PC+1];
      address = ((unsigned int)mem[address+1] << 8) + (mem[address]+regs->YR);
      x = mem[address];
     	regs->PC += 2;
      break;

	}


	if (address == 56321)
	{
		mem[56321] = getkey(mem[56320]);
		x = mem[56321];
	}

	switch  (code)
	{
		case OP_CMP:
			help = regs->AC-x;
      regs->Zero_Flag =  help==0;
      regs->Negative_Flag = (help & 128);
      regs->Carry_Flag = help < 256;
  		break;

		case OP_CPX:
			help = regs->XR-x;
      regs->Zero_Flag =  help==0;
      regs->Negative_Flag = (help & 128);
      regs->Carry_Flag = help < 256;
  		break;

		case OP_CPY:
			help = regs->YR-x;
      regs->Zero_Flag =  help==0;
      regs->Negative_Flag = (help & 128);
      regs->Carry_Flag = help < 256;
  		break;

		case OP_LDA:
      regs->AC = x;
      regs->Zero_Flag =  x==0;
      regs->Negative_Flag = (x & 128);
      break;

		case OP_STA:
      x = regs->AC;
      break;

		case OP_LDX:
      regs->XR = x;
      regs->Zero_Flag =  x==0;
      regs->Negative_Flag = (x & 128);
      break;

		case OP_STX:
      x = regs->XR;
      break;

		case OP_LDY:
      regs->YR = x;
      regs->Zero_Flag =  x==0;
      regs->Negative_Flag = (x & 128) ;
      break;

		case OP_STY:
      x = regs->YR;
      break;

		case OP_AND:
			regs->AC = regs->AC & x;
      regs->Zero_Flag =  regs->AC==0;
      regs->Negative_Flag = (regs->AC & 128) ;
      break;

		case OP_EOR:
			regs->AC = regs->AC ^ x;
      regs->Zero_Flag =  regs->AC==0;
      regs->Negative_Flag = (regs->AC & 128) ;
      break;

		case OP_ORA:
			regs->AC = regs->AC | x;
      regs->Zero_Flag =  regs->AC==0;
      regs->Negative_Flag = (regs->AC & 128) ;
      break;

		case OP_ADC:
			help2 = x;
			if (regs->Carry_Flag) {help = regs->AC+x+1;} else {help = regs->AC+x;}
			regs->Overflow_Flag = 	(((~(regs->AC ^ help2) & 0x80) )) && 
														(((regs->AC ^ help) & 0x80) );
      regs->AC = help & 255;
      regs->Zero_Flag =  regs->AC==0;
      regs->Negative_Flag = (regs->AC & 128) ;
      regs->Carry_Flag = help > 255;
      break;

		case OP_SBC:
			help2 = x;
			if (regs->Carry_Flag) {help = regs->AC-x;} else {help = regs->AC-x-1;}
			regs->Overflow_Flag = 	((((regs->AC ^ help) & 0x80) )) && 
														(((regs->AC ^ help) & 0x80) );

      regs->AC = help & 255;
      regs->Carry_Flag = help < 256;
      regs->Zero_Flag =  regs->AC==0;
      regs->Negative_Flag = (regs->AC & 128) ;
      break;

		case OP_BIT:
      regs->Negative_Flag = (x & 128) ;
      regs->Overflow_Flag = (x & 64) ;
      regs->Zero_Flag =  !(x & regs->AC );
      break;

		case OP_INC:
      x++;
      regs->Zero_Flag =  x==0;
      regs->Negative_Flag = (x & 128) ;
      break;

		case OP_INX:
      regs->XR++;
      regs->Zero_Flag =  regs->XR==0;
      regs->Negative_Flag = (regs->XR & 128) ;
      break;

		case OP_INY:
      regs->YR++;
      regs->Zero_Flag =  regs->YR==0;
      regs->Negative_Flag = (regs->YR & 128) ;
      break;
      
		case OP_DEC:
      x--;
      regs->Zero_Flag =  x==0;
      regs->Negative_Flag = (x & 128) ;
      break;

		case OP_DEX:
      regs->XR--;
      regs->Zero_Flag =  regs->XR==0;
      regs->Negative_Flag = (regs->XR & 128) ;
      break;

		case OP_DEY:
      regs->YR--;
      regs->Zero_Flag =  regs->YR==0;
      regs->Negative_Flag = (regs->YR & 128) ;
      break;

		case OP_ASL:
      regs->Carry_Flag = (x & 128) ;
			x = x << 1;
      regs->Zero_Flag =  x==0;
      regs->Negative_Flag = (x & 128) ;
      break;

		case OP_LSR:
      regs->Carry_Flag = (x & 1) ;
			x = x >> 1;
      regs->Zero_Flag =  x==0;
      regs->Negative_Flag = 0;
      break;

		case OP_ROL:
      help = x << 1;
      if (regs->Carry_Flag) help = help | 1;
			regs->Carry_Flag = help > 255;			
			x = help & 255;
      regs->Zero_Flag =  x==0;
      regs->Negative_Flag = (x & 128) ;
      break;

		case OP_ROR:
      help = 0;
      if ((x & 1) == 1) help = 1;
			x = x >> 1;
      if (regs->Carry_Flag) x = x | 128;
      regs->Carry_Flag = (help & 1) == 1;
      regs->Zero_Flag =  x==0;
      regs->Negative_Flag = (x & 128) ;
      break;

		case OP_PHA:
			mem[256+regs->SP] = regs->AC;
			regs->SP--;
      break;

		case OP_PLA:
			regs->SP++;
			regs->AC = mem[256+regs->SP];
      regs->Zero_Flag = regs->AC==0;
      regs->Negative_Flag = (regs->AC & 128) ;
      break;

		case OP_PHP:
			mem[256+regs->SP] = getstatus() | 16;
			regs->SP--;
      break;

		case OP_PLP:
			regs->SP++;
			stack = mem[256+regs->SP];
			putstatus(stack);
      break;

		case OP_JMP:
      regs->PC = address;
      break;

		case OP_JSR:
			mem[256+regs->SP] = ((regs->PC-1) >> 8) & 0xFF;
			regs->SP--;
			mem[256+regs->SP] = (regs->PC-1) & 0xFF;
			regs->SP--;
      regs->PC = address;
      break;

		case OP_RTS:
			regs->SP++;
			regs->PC = mem[256+regs->SP];
			regs->SP++;
			regs->PC = regs->PC | ((unsigned int)mem[256+regs->SP] << 8);
      regs->PC++;
      break;

		case OP_RTI:
			regs->SP++;
			stack = mem[256+regs->SP];
			putstatus(stack);
			regs->SP++;
			regs->PC = mem[256+regs->SP];
			regs->SP++;
			regs->PC = regs->PC | ((unsigned int)mem[256+regs->SP] << 8);
      break;

		case OP_BRK:
			mem[256+regs->SP] = ((regs->PC) >> 8) & 0xFF;
			regs->SP--;
			mem[256+regs->SP] = (regs->PC) & 0xFF;
			regs->SP--;

			mem[256+regs->SP] = getstatus();
			regs->SP--;
			regs->IRQ_Disable_Flag = 1;
			regs->Brk_Flag = 1;
      regs->PC = (unsigned int)mem[0xFFFE] | ((unsigned int)mem[0xFFFF] << 8);
      break;

		case OP_BNE:
      if (!regs->Zero_Flag) regs->PC += (signed char)x;
      break;

		case OP_BCC:
      if (!regs->Carry_Flag) regs->PC += (signed char)x;
      break;

		case OP_BEQ:
      if (regs->Zero_Flag) regs->PC += (signed char)x;
      break;

		case OP_BMI:
      if (regs->Negative_Flag) regs->PC += (signed char)x;
      break;

		case OP_BVC:
      if (regs->Overflow_Flag) regs->PC += (signed char)x;
      break;

		case OP_BCS:
      if (regs->Carry_Flag) regs->PC += (signed char)x;
      break;

		case OP_BPL:
      if (!regs->Negative_Flag) regs->PC += (signed char)x;
      break;

		case OP_BVS:
      if (regs->Overflow_Flag) regs->PC += (signed char)x;
      break;

		case OP_TAX:
      regs->XR = regs->AC;
      regs->Zero_Flag =  regs->AC==0;
      regs->Negative_Flag = (regs->AC & 128) ;
      break;

		case OP_TAY:
      regs->YR = regs->AC;
      regs->Zero_Flag =  regs->AC==0;
      regs->Negative_Flag = (regs->AC & 128) ;
      break;

		case OP_TXA:
      regs->AC = regs->XR;
      regs->Zero_Flag =  regs->AC==0;
      regs->Negative_Flag = (regs->AC & 128) ;
      break;

		case OP_TYA:
      regs->AC = regs->YR;
      regs->Zero_Flag =  regs->AC==0;
      regs->Negative_Flag = (regs->AC & 128) ;
      break;

		case OP_TXS:
      regs->SP = regs->XR;
      break;

		case OP_TSX:
      regs->XR = regs->SP;
      regs->Zero_Flag =  regs->XR==0;
      regs->Negative_Flag = (regs->XR & 128) ;
      break;

		case OP_CLC:
      regs->Carry_Flag = 0;
      break;

		case OP_CLD:
      regs->Decimal_Flag = 0;
      break;

		case OP_CLI:
      regs->IRQ_Disable_Flag = 0;
      break;

		case OP_CLV:
      regs->Overflow_Flag = 0;
      break;

		case OP_SEC:
      regs->Carry_Flag = 1;
      break;

		case OP_SED:
      regs->Decimal_Flag = 1;
      break;

		case OP_SEI:
      regs->IRQ_Disable_Flag = 1;
      break;

	}

	switch (address_mode)
	{
		case AMODE_ACC:
			regs->AC = x;
      break;
		case AMODE_IMM:
      break;
		case AMODE_IMP:
      break;
		case AMODE_REL:
      break;
		case AMODE_IND:
      break;

		default:
			mem[address] = x;
			break;
	}

}

